import os
import logging
from typing import Union

import rasterio
import geopandas as gpd


def create_basin_dem(basin_shp: str,
                     dem_file: str,
                     run_name: str,
                     output_directory: Union[str, None] = None) -> str:
    """Mask the input DEM using a basin geometry representative of the contributing area.

    :param basin_shp:               Full path with file name and extension to the target basin shapefile
    :type basin_shp:                str

    :param dem_file:                Full path with file name and extension to the input DEM raster file.
    :type dem_file:                 str

    :param run_name:                Name of run, all lowercase and only underscore separated.
    :type run_name:                 str

    :param output_directory:        Full path to a write-enabled directory to write output files to
    :type output_directory:         Union[str, None]

    :return:                        Full path with file name and extension to the masked DEM raster file

    """

    # dissolve target basin geometries
    basin_geom = gpd.read_file(basin_shp).dissolve().geometry.values[0]

    with rasterio.open(dem_file) as src:
        if src.crs is None:
            logging.warning("Input DEM raster does not have a defined coordinate reference system.")

        # apply basin geometry as a mask
        out_image, out_transform = rasterio.mask.mask(src, basin_geom, crop=True)

        # update the raster metadata with newly cropped extent
        out_meta = src.meta
        out_meta.update({"driver": "GTiff",
                         "height": out_image.shape[1],
                         "width": out_image.shape[2],
                         "transform": out_transform})

        # write outputs
        output_file = os.path.join(output_directory, f"dem_masked_{run_name}.tif")
        with rasterio.open(output_file, "w", **out_meta) as dest:
            dest.write(out_image)

        return output_file



import rasterio
import numpy as np
import pandas as pd

from typing import Union


def hypsometric_curve(dem_file: str,
                      elevation_interval: float = 0.1,
                      min_elevation: Union[float, None] = None,
                      max_elevation: Union[float, None] = None) -> pd.DataFrame:
    """Calculate a hypsometric curve as an elevation-area relationship assessment metric
    of the landform shape at a site.  Provides basic metric of opportunity for inundation and
    habitat opportunity.

    :param dem_file:                Full path with file name and extension to the input digital elevation model
                                    raster
    :type dem_file:                 str

    :param elevation_interval:      Elevation sample spacing in the units of the input DEM
    :type elevation_interval:       float

    :param min_elevation:           Optional.  Minimum elevation to sample from.  Default is to use the minimum
                                    elevation of the raster as a starting point.
    :type min_elevation:            float

    :param max_elevation:           Optional.  Maximum elevation to sample from.  Default is to use the maximum
                                    elevation of the raster as a starting point.
    :type max_elevation:            float

    :return:                        Pandas DataFrame of elevation, area at or above the target elevation, and
                                    percent area at or above the target elevation for each elevation interval

    """

    # create a dictionary to hold results
    result_dict = {'dem_elevation': [], 'dem_area_at_elevation': []}

    with rasterio.open(dem_file) as src:

        # read the raster band into a number array
        arr = src.read(1)

        # convert the raster nodata value to numpy nan
        arr[arr == src.nodata] = np.nan

        # grid cell resolution
        grid_cell_area = np.abs(src.transform[0] * src.transform[4])

        # set minimum and maximum elevation value; use raster determined values by default
        if min_elevation is None:
            min_elevation = np.nanmin(arr)

        if max_elevation is None:
            max_elevation = np.nanmax(arr)

        # create elevation intervals to process
        elev_slices = np.arange(min_elevation, max_elevation + elevation_interval, elevation_interval)

        # calculate each area at elevation intervals
        for i in elev_slices:
            result_dict['dem_elevation'].append(i)
            result_dict['dem_area_at_elevation'].append(np.where(arr >= i)[0].shape[0] * grid_cell_area)

        # convert results to data frame
        df = pd.DataFrame(result_dict)

        # calculate the total area
        total_area = df['dem_area_at_elevation'].max()

        # calculate percent area per elevation slice
        df['dem_percent_area'] = df['dem_area_at_elevation'] / total_area

        return df



import logging

import pandas as pd


def import_gage_data(gage_data_file: str,
                      date_field_name: str = 'DATE',
                      time_field_name: str = 'TIME'):
    """Process gage data tabular file.

    :param gage_data_file:          Full path with file name and extension to the gage data file.
    :type gage_data_file:           str

    :param date_field_name:         Name of date field in file
    :type date_field_name:          str

    :param time_field_name:         Name of time field in file
    :type time_field_name:          str

    :return:                        DataFrame of time series gage data sorted by date_time

    """

    # convert date and time strings to a pandas datetime type
    df = pd.read_csv(gage_data_file,  infer_datetime_format=True, parse_dates=[[date_field_name, time_field_name]])

    # rename date time field
    df.rename(columns={f'{date_field_name}_{time_field_name}': 'date_time'}, inplace=True)

    return df.sort_values(by=['date_time'])


def process_gage_data(gage_data_file: str,
                      date_field_name: str = 'DATE',
                      time_field_name: str = 'TIME',
                      elevation_field_name: str = 'WL_ELEV_M',
                      verbose: bool = False):
    """Process gage data tabular file.

    :param gage_data_file:          Full path with file name and extension to the gage data file.
    :type gage_data_file:           str

    :param date_field_name:         Name of date field in file
    :type date_field_name:          str

    :param time_field_name:         Name of time field in file
    :type time_field_name:          str

    :param elevation_field_name:    Name of elevation field in file
    :type elevation_field_name:     str

    :param verbose:                 Choice to log verbose description of file
    :type verbose:                  bool

    :returns:                       [0] minumum water elevation in file
                                    [1] maximum water elevation in file
                                    [2] dictionary of water elevation frequency {elev:  frequency}
    """

    # import gage data as time series
    df = import_gage_data(gage_data_file=gage_data_file,
                          date_field_name=date_field_name,
                          time_field_name=time_field_name)

    min_wtr_elev = df[elevation_field_name].min()
    max_wtr_elev = df[elevation_field_name].max()
    d_freq = df[elevation_field_name].value_counts().to_dict()

    if verbose:

        # calculate the number of days in the file
        n_days = (df['date_time'].max() - df['date_time'].min()).days

        logging.info(f"Total Time Steps:  {df.shape[0]}")
        logging.info(f"Days Verification:  {n_days}")
        logging.info(f"Minimum Water Surface Elevation:  {min_wtr_elev}")
        logging.info(f"Maximum Water Surface Elevation:  {max_wtr_elev}")

    return min_wtr_elev, max_wtr_elev, d_freq



from typing import Tuple
from typing import Union

import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
from scipy import stats
from matplotlib.ticker import ScalarFormatter, FormatStrFormatter, FixedLocator

from .gage import import_gage_data
from .inflection import calculate_bankfull_elevation
from .hypsometric import hypsometric_curve


class PlotGeneral:
    """General plotting setup and option handling.

    :param style:                   Seaborn style designation
    :type style:                    str

    :param font_scale:              Scaling factor for font size
    :type font_scale:               float

    :param figsize:                 Tuple of figure size (x, y)
    :type figsize:                  Tuple[int]

    :param show_plot:               If True, plot will be displayed
    :type show_plot:                bool

    :param save_plot:               If True, plot will be written to file and a value must be set for output_file
    :type save_plot:                bool

    :param output_file:             Full path with file name and extension to an output file
    :type output_file:              str

    :param dpi:                     The resolution in dots per inch
    :type dpi:                      int

    """

    def __init__(self,
                 style: str = 'whitegrid',
                 font_scale: float = 1.2,
                 figsize: Tuple[int] = (12, 8),
                 show_plot: bool = True,
                 save_plot: bool = False,
                 output_file: Union[str, None] = None,
                 dpi: int = 150):

        # set up plot style
        sns.set(style=style, font_scale=font_scale)

        # setup figure and axis
        self.fig, self.ax = plt.subplots(figsize=figsize)

        self.show_plot = show_plot
        self.save_plot = save_plot
        self.output_file = output_file
        self.dpi = dpi

    def output_handler(self):
        """Handle plot display and output options."""

        # save figure
        if self.save_plot:

            # ensure a value is set for output_file
            if self.output_file is None:
                raise AssertionError("If writing plot to file, you must set a value for 'output_file'")

            plt.savefig(self.output_file, dpi=self.dpi)

        # show plot
        if self.show_plot:
            plt.show()

    def set_xaxis_to_thousands(self):
        """Format the x-axis to show thousands as K (e.g., 15,000 as 15K)."""

        # format the x-axis to show thousands as K (e.g., 15,000 as 15K)
        self.ax.xaxis.set_major_formatter(
            ticker.FuncFormatter(lambda x, pos: '{:}'.format(int(x / 1000)) + 'K' if x >= 1000 else int(x))
        )


class PlotGageData(PlotGeneral):
    """Prepare gage data for plotting.

    :param gage_data:               Either a full path with file name and extension to the gage data file or
                                    a data frame processed using atiim.import_gage_data()
    :type gage_data:                Union[str, pd.Dataframe]

    :param date_field_name:         Name of date field in file
    :type date_field_name:          str

    :param time_field_name:         Name of time field in file
    :type time_field_name:          str

    :param elevation_field_name:    Name of elevation field in file
    :type elevation_field_name:     str

    :param style:                   Seaborn style designation
    :type style:                    str

    :param font_scale:              Scaling factor for font size
    :type font_scale:               float

    :param figsize:                 Tuple of figure size (x, y)
    :type figsize:                  Tuple[int]

    :param show_plot:               If True, plot will be displayed
    :type show_plot:                bool

    :param save_plot:               If True, plot will be written to file and a value must be set for output_file
    :type save_plot:                bool

    :param output_file:             Full path with file name and extension to an output file
    :type output_file:              str

    :param dpi:                     The resolution in dots per inch
    :type dpi:                      int

    """

    def __init__(self,
                 data: Union[str, pd.DataFrame],
                 date_field_name: str = 'DATE',
                 time_field_name: str = 'TIME',
                 elevation_field_name: str = 'WL_ELEV_M',
                 style: str = 'whitegrid',
                 font_scale: float = 1.2,
                 figsize: Tuple[int] = (12, 8),
                 show_plot: bool = True,
                 save_plot: bool = False,
                 output_file: Union[str, None] = None,
                 dpi: int = 150):

        # initialize parent class
        super().__init__(style=style,
                         font_scale=font_scale,
                         figsize=figsize,
                         show_plot=show_plot,
                         save_plot=save_plot,
                         output_file=output_file,
                         dpi=dpi)

        self.data = data
        self.date_field_name = date_field_name
        self.time_field_name = time_field_name
        self.elevation_field_name = elevation_field_name

        # setup data frame
        self.df = self.prepare_data()

        self.z_max = self.df[elevation_field_name].max()
        self.z_min = self.df[elevation_field_name].min()
        self.n_records = self.df.shape[0]

        # sort elevation values in ascending order
        self.z_sorted = self.sorted_elevation()

        # set cumulative y-axis values
        self.y_cumulative_series = self.set_cumulative_series()

        # calculate lognormal fit
        self.shape, self.location, self.scale = self.calculate_lognorm_fit()

    def prepare_data(self) -> pd.DataFrame:
        """Assign gage data to a data frame."""

        if type(self.data) == pd.DataFrame:
            return self.data

        else:
            return import_gage_data(self.data, self.date_field_name, self.time_field_name)

    def generate_wse_interval(self, x_padding: float = 1.05, n_samples: int = 100) -> np.ndarray:
        """Generate an evenly spaced interval of values encompassing water surface elevation
        for use in scaling the x-axis.

        :param x_padding:               Multiplier for maximum elevation to determine an ending interval for the x-axis.
                                        E.g., if max value is 100 and x_padding is 1.1 then the ending bound would be 110.
        :type x_padding:                float

        :param n_samples:               The number of samples to generate over the x-axis space
        :type n_samples:                int

        :return:                        NumPy array of evenly spaced water surface elevation intervals

        """

        return np.linspace(self.df[self.elevation_field_name].min(),
                           self.df[self.elevation_field_name].max() * x_padding,
                           num=n_samples)

    def sorted_elevation(self) -> pd.Series:
        """Sort elevation data in ascending order."""

        return self.df[self.elevation_field_name].sort_values()

    def calculate_lognorm_fit(self) -> Tuple[float]:
        """Calculate the lognormal continuous random variable and generate parameter estimates. Uses
        Maximum Likelihood Estimateion (MLE).

        :returns:                               [0] shape parameter estimates
                                                [1] location parameters
                                                [2] scale parameters
        """

        return stats.lognorm.fit(self.z_sorted)

    def calculate_cdf(self, data: np.ndarray) -> np.ndarray:
        """Calculate the cumulative distribution.

        :param data:                    Input data for x-axis values
        :type data:                     np.ndarray

        :return:                        NumPy array of CDF outputs

        """

        return stats.lognorm.cdf(data, self.shape, self.location, self.scale)

    def calculate_pdf(self, data: np.ndarray) -> np.ndarray:
        """Calculate the probability density.

        :param data:                    Input data for x-axis values
        :type data:                     np.ndarray

        :return:                        NumPy array of PDF outputs

        """

        return stats.lognorm.pdf(data, self.shape, self.location, self.scale)

    def set_cumulative_series(self) -> pd.Series:
        """Generate axis series and intervals for the cumulative distribution y-axis."""

        # generate axis bounds and intervals for the cumulative distribution y-axis
        y_cumulative = np.linspace(0.0, 1.0, self.n_records)

        return pd.Series(y_cumulative, index=self.z_sorted)

    def calculate_exceedance_probability(self) -> pd.Series:
        """Calculate a y-axis series of exceedance probilites as a percentage."""

        return (1.0 - self.y_cumulative_series) * 100


class PlotInundationData(PlotGeneral):
    """Prepare inundation data for plotting.

    :param data:                    Either a full path with file name and extension to the inundation data file or
                                    a data frame processed using atiim.simulate_inundation()
    :type data:                     Union[str, pd.Dataframe]

    :param style:                   Seaborn style designation
    :type style:                    str

    :param font_scale:              Scaling factor for font size
    :type font_scale:               float

    :param figsize:                 Tuple of figure size (x, y)
    :type figsize:                  Tuple[int]

    :param show_plot:               If True, plot will be displayed
    :type show_plot:                bool

    :param save_plot:               If True, plot will be written to file and a value must be set for output_file
    :type save_plot:                bool

    :param output_file:             Full path with file name and extension to an output file
    :type output_file:              str

    :param dpi:                     The resolution in dots per inch
    :type dpi:                      int

    """

    # field names in inundation standard output
    FLD_ID = 'id'
    FLD_FREQUENCY = 'frequency'
    FLD_ELEVATION = 'elevation'
    FLD_AREA = 'area'
    FLD_HECTARES = 'hectares'
    FLD_PERIMETER = 'perimeter'
    FLD_HECTARE_HOURS = 'hect_hours'

    def __init__(self,
                 data: Union[str, pd.DataFrame],
                 style: str = 'whitegrid',
                 font_scale: float = 1.2,
                 figsize: Tuple[int] = (12, 8),
                 show_plot: bool = True,
                 save_plot: bool = False,
                 output_file: Union[str, None] = None,
                 dpi: int = 150):

        # initialize parent class
        super().__init__(style=style,
                         font_scale=font_scale,
                         figsize=figsize,
                         show_plot=show_plot,
                         save_plot=save_plot,
                         output_file=output_file,
                         dpi=dpi)

        self.data = data

        # read data in as data frame
        self.df = self.prepare_data()

        # unpack fields
        self.id = self.df[PlotInundationData.FLD_ID]
        self.frequency = self.df[PlotInundationData.FLD_FREQUENCY]
        self.elevation = self.df[PlotInundationData.FLD_ELEVATION]
        self.area = self.df[PlotInundationData.FLD_AREA]
        self.hectares = self.df[PlotInundationData.FLD_HECTARES]
        self.perimeter = self.df[PlotInundationData.FLD_PERIMETER]
        self.hectare_hours = self.df[PlotInundationData.FLD_HECTARE_HOURS]

    def prepare_data(self) -> pd.DataFrame:
        """Assign input data to a data frame."""

        if type(self.data) == pd.DataFrame:
            return self.data

        else:
            return pd.read_csv(self.data)


class PlotDem(PlotGeneral):
    """Prepare elevation data from an input DEM raster for plotting.

    :param data:                    Either a full path with file name and extension to the input DEM raster or
                                    a data frame processed using atiim.hypsometric_curve()
    :type data:                     Union[str, pd.Dataframe]

    :param elevation_interval:      Optional. Elevation sample spacing in the units of the input DEM. Set when using a
                                    DEM raster file input.
    :type elevation_interval:       float

    :param min_elevation:           Optional.  Minimum elevation to sample from.  Default is to use the minimum
                                    elevation of the raster as a starting point. Set when using a DEM raster file input.
    :type min_elevation:            float

    :param max_elevation:           Optional.  Maximum elevation to sample from.  Default is to use the maximum
                                    elevation of the raster as a starting point. Set when using a DEM raster file input.
    :type max_elevation:            float

    :param style:                   Seaborn style designation
    :type style:                    str

    :param font_scale:              Scaling factor for font size
    :type font_scale:               float

    :param figsize:                 Tuple of figure size (x, y)
    :type figsize:                  Tuple[int]

    :param show_plot:               If True, plot will be displayed
    :type show_plot:                bool

    :param save_plot:               If True, plot will be written to file and a value must be set for output_file
    :type save_plot:                bool

    :param output_file:             Full path with file name and extension to an output file
    :type output_file:              str

    :param dpi:                     The resolution in dots per inch
    :type dpi:                      int

    """

    def __init__(self,
                 data: Union[str, pd.DataFrame],
                 elevation_interval: float,
                 min_elevation: Union[float, None] = None,
                 max_elevation: Union[float, None] = None,
                 style: str = 'whitegrid',
                 font_scale: float = 1.2,
                 figsize: Tuple[int] = (12, 8),
                 show_plot: bool = True,
                 save_plot: bool = False,
                 output_file: Union[str, None] = None,
                 dpi: int = 150):

        # initialize parent class
        super().__init__(style=style,
                         font_scale=font_scale,
                         figsize=figsize,
                         show_plot=show_plot,
                         save_plot=save_plot,
                         output_file=output_file,
                         dpi=dpi)

        self.data = data
        self.elevation_interval = elevation_interval
        self.min_elevation = min_elevation
        self.max_elevation = max_elevation

        # read data in as data frame
        self.df = self.prepare_data()

    def prepare_data(self) -> pd.DataFrame:
        """Assign input data to a data frame."""

        if type(self.data) == pd.DataFrame:
            return self.data

        else:
            return hypsometric_curve(dem_file=self.data,
                                     elevation_interval=self.elevation_interval,
                                     min_elevation=self.min_elevation,
                                     max_elevation=self.max_elevation)


def plot_wse_timeseries(data: Union[str, pd.DataFrame],
                        show_plot: bool = True,
                        save_plot: bool = False,
                        output_file: Union[str, None] = None,
                        dpi: int = 150,
                        date_field_name: str = 'DATE',
                        time_field_name: str = 'TIME',
                        elevation_field_name: str = 'WL_ELEV_M',
                        style: str = 'whitegrid',
                        font_scale: float = 1.2,
                        figsize: Tuple[int] = (12, 8),
                        color: str = '#005AB5',
                        transparency: float = 0.7,
                        title: Union[str, None] = 'Water Surface Elevation Gage Measurements'):
    """Create plot for water surface elevation for the gage measurement period.

    :param data:                    Either a full path with file name and extension to the gage data file or
                                    a data frame processed using atiim.import_gage_data()
    :type data:                     Union[str, pd.Dataframe]

    :param show_plot:               If True, plot will be displayed
    :type show_plot:                bool

    :param save_plot:               If True, plot will be written to file and a value must be set for output_file
    :type save_plot:                bool

    :param output_file:             Full path with file name and extension to an output file
    :type output_file:              str

    :param dpi:                     The resolution in dots per inch
    :type dpi:                      int

    :param date_field_name:         Name of date field in file
    :type date_field_name:          str

    :param time_field_name:         Name of time field in file
    :type time_field_name:          str

    :param elevation_field_name:    Name of elevation field in file
    :type elevation_field_name:     str

    :param style:                   Seaborn style designation
    :type style:                    str

    :param font_scale:              Scaling factor for font size
    :type font_scale:               float

    :param figsize:                 Tuple of figure size (x, y)
    :type figsize:                  Tuple[int]

    :param color:                   Color of line and markers in plot
    :type color:                    str

    :param transparency             Alpha value from 0 to 1 for transparency
    :type transparency              float

    :param title                    Plot title.  If None, then no title will be displayed.
    :type title                     Union[str, None]

    """

    # prepare data for plotting
    plot = PlotGageData(data=data,
                        date_field_name=date_field_name,
                        time_field_name=time_field_name,
                        elevation_field_name=elevation_field_name,
                        style=style,
                        font_scale=font_scale,
                        figsize=figsize,
                        show_plot=show_plot,
                        save_plot=save_plot,
                        output_file=output_file,
                        dpi=dpi)

    # construct a line plot
    sns.lineplot(x="date_time",
                 y=elevation_field_name,
                 data=plot.df,
                 color=color,
                 alpha=transparency)

    plot.ax.set(ylabel='Water Surface Elevation (m)',
                xlabel=None,
                title=title)

    # set limits for the x-axis
    plt.xlim(xmin=plot.df['date_time'].min(), xmax=plot.df['date_time'].max())

    plt.xticks(rotation=45)

    # handle display or output options
    plot.output_handler()

    plt.close()


def plot_wse_cumulative_distribution(data: Union[str, pd.DataFrame],
                                     show_plot: bool = True,
                                     save_plot: bool = False,
                                     output_file: Union[str, None] = None,
                                     dpi: int = 150,
                                     date_field_name: str = 'DATE',
                                     time_field_name: str = 'TIME',
                                     elevation_field_name: str = 'WL_ELEV_M',
                                     x_padding: float = 1.05,
                                     n_samples: int = 100,
                                     style: str = 'whitegrid',
                                     font_scale: float = 1.2,
                                     figsize: Tuple[int] = (12, 8),
                                     data_color: str = '#005AB5',
                                     lognorm_color: str = '#DC3220',
                                     title: Union[str, None] = 'Cumulative Distribution of Water Surface Elevation'):
    """Plot the cumulative distribution function for water surface elevation from the gage data.

    :param data:                     Either a full path with file name and extension to the gage data file or
                                    a data frame processed using atiim.import_gage_data()
    :type data:                     Union[str, pd.Dataframe]

    :param show_plot:               If True, plot will be displayed
    :type show_plot:                bool

    :param save_plot:               If True, plot will be written to file and a value must be set for output_file
    :type save_plot:                bool

    :param date_field_name:         Name of date field in file
    :type date_field_name:          str

    :param time_field_name:         Name of time field in file
    :type time_field_name:          str

    :param elevation_field_name:    Name of elevation field in file
    :type elevation_field_name:     str

    :param x_padding:               Multiplier for maximum elevation to determine an ending interval for the x-axis.
                                    E.g., if max value is 100 and x_padding is 1.1 then the ending bound would be 110.
    :type x_padding:                float

    :param n_samples:               The number of samples to generate over the x-axis space
    :type n_samples:                int

    :param output_file:             Full path with file name and extension to an output file
    :type output_file:              str

    :param dpi:                     The resolution in dots per inch
    :type dpi:                      int

    :param style:                   Seaborn style designation
    :type style:                    str

    :param font_scale:              Scaling factor for font size
    :type font_scale:               float

    :param figsize:                 Tuple of figure size (x, y)
    :type figsize:                  Tuple[int]

    :param data_color:              Color of data line
    :type data_color:               str

    :param lognorm_color:           Color of data line
    :type lognorm_color:            str

    :param title                    Plot title.  If None, then no title will be displayed.
    :type title                     Union[str, None]

    :param title                    Plot title.  If None, then no title will be displayed.
    :type title                     Union[str, None]

    """

    # prepare data for plotting
    plot = PlotGageData(data=data,
                        date_field_name=date_field_name,
                        time_field_name=time_field_name,
                        elevation_field_name=elevation_field_name,
                        style=style,
                        font_scale=font_scale,
                        figsize=figsize,
                        show_plot=show_plot,
                        save_plot=save_plot,
                        output_file=output_file,
                        dpi=dpi)

    # generate an evenly spaced interval of values encompassing water surface elevation for use in scaling the x-axis
    x_data = plot.generate_wse_interval(x_padding=x_padding, n_samples=n_samples)

    # plot elevation data series steps
    plot.y_cumulative_series.plot(ax=plot.ax, drawstyle='steps', label='data', color=data_color)

    # plot lognormal curve
    plot.ax.plot(x_data, plot.calculate_cdf(data=x_data), label='lognormal', color=lognorm_color)

    plot.ax.set_xlabel('Water Surface Elevation (m)')
    plot.ax.set_ylabel('Cumulative Distribution')
    plot.ax.legend(loc=0, framealpha=0.5)
    plt.title(title)

    # set x-axis limits
    plt.xlim(xmin=x_data.min(), xmax=plot.z_max)

    # handle display or output options
    plot.output_handler()

    plt.close()


def plot_wse_probability_density(data: Union[str, pd.DataFrame],
                                 show_plot: bool = True,
                                 save_plot: bool = False,
                                 output_file: Union[str, None] = None,
                                 dpi: int = 150,
                                 date_field_name: str = 'DATE',
                                 time_field_name: str = 'TIME',
                                 elevation_field_name: str = 'WL_ELEV_M',
                                 x_padding: float = 1.05,
                                 n_samples: int = 100,
                                 style: str = 'whitegrid',
                                 font_scale: float = 1.2,
                                 figsize: Tuple[int] = (12, 8),
                                 data_color: str = '#005AB5',
                                 lognorm_color: str = '#DC3220',
                                 transparency: float = 0.6,
                                 n_bins: int = 21,
                                 title: Union[str, None] = 'Probability Density of Water Surface Elevation'):
    """Plot the probability density of water surface elevations based on the period-of-record.

    :param data:                    Either a full path with file name and extension to the gage data file or
                                    a data frame processed using atiim.import_gage_data()
    :type data:                     Union[str, pd.Dataframe]

    :param show_plot:               If True, plot will be displayed
    :type show_plot:                bool

    :param save_plot:               If True, plot will be written to file and a value must be set for output_file
    :type save_plot:                bool

    :param date_field_name:         Name of date field in file
    :type date_field_name:          str

    :param time_field_name:         Name of time field in file
    :type time_field_name:          str

    :param elevation_field_name:    Name of elevation field in file
    :type elevation_field_name:     str

    :param x_padding:               Multiplier for maximum elevation to determine an ending interval for the x-axis.
                                    E.g., if max value is 100 and x_padding is 1.1 then the ending bound would be 110.
    :type x_padding:                float

    :param n_samples:               The number of samples to generate over the x-axis space
    :type n_samples:                int

    :param output_file:             Full path with file name and extension to an output file
    :type output_file:              str

    :param dpi:                     The resolution in dots per inch
    :type dpi:                      int

    :param style:                   Seaborn style designation
    :type style:                    str

    :param font_scale:              Scaling factor for font size
    :type font_scale:               float

    :param figsize:                 Tuple of figure size (x, y)
    :type figsize:                  Tuple[int]

    :param data_color:              Color of data line
    :type data_color:               str

    :param lognorm_color:           Color of data line
    :type lognorm_color:            str

    :param transparency             Alpha value from 0 to 1 for transparency of histogram
    :type transparency              float

    :param n_bins:                  The number of equal-width bins in the range.
    :type n_bins:                   int

    :param title                    Plot title.  If None, then no title will be displayed.
    :type title                     Union[str, None]

    """

    # prepare data for plotting
    plot = PlotGageData(data=data,
                        date_field_name=date_field_name,
                        time_field_name=time_field_name,
                        elevation_field_name=elevation_field_name,
                        style=style,
                        font_scale=font_scale,
                        figsize=figsize,
                        show_plot=show_plot,
                        save_plot=save_plot,
                        output_file=output_file,
                        dpi=dpi)

    # sort elevation value in ascending order
    z_sorted = plot.sorted_elevation()

    # generate an evenly spaced interval of values encompassing water surface elevation for use in scaling the x-axis
    x_data = plot.generate_wse_interval(x_padding=x_padding, n_samples=n_samples)

    # plot binned elevation data
    plot.ax.hist(z_sorted, bins=n_bins, label='data', density=True, color=data_color, alpha=transparency)

    # plot PDF outputs
    pdf = plot.calculate_pdf(data=x_data)
    plot.ax.plot(x_data, pdf, label='lognormal', color=lognorm_color)

    plot.ax.set_xlabel('Water Surface Elevation (m)')
    plot.ax.set_ylabel('Probability Density')
    plot.ax.legend(loc=0, framealpha=0.5)
    plt.title(title)

    # set x-axis limits
    plt.xlim(xmin=x_data.min(), xmax=x_data.max())

    # handle display or output options
    plot.output_handler()

    plt.close()


def plot_wse_exceedance_probability(data: Union[str, pd.DataFrame],
                                    show_plot: bool = True,
                                    save_plot: bool = False,
                                    output_file: Union[str, None] = None,
                                    dpi: int = 150,
                                    date_field_name: str = 'DATE',
                                    time_field_name: str = 'TIME',
                                    elevation_field_name: str = 'WL_ELEV_M',
                                    x_padding: float = 1.05,
                                    n_samples: int = 100,
                                    style: str = 'whitegrid',
                                    font_scale: float = 1.2,
                                    figsize: Tuple[int] = (12, 8),
                                    data_color: str = '#005AB5',
                                    lognorm_color: str = '#DC3220',
                                    title: Union[str, None] = 'Exceedance Probability of Water Surface Elevation'):
    """Plot the probability of occurrence for individual water surface elevations based on the period-of-record.

    :param data:                    Either a full path with file name and extension to the gage data file or
                                    a data frame processed using atiim.import_gage_data()
    :type data:                     Union[str, pd.Dataframe]

    :param show_plot:               If True, plot will be displayed
    :type show_plot:                bool

    :param save_plot:               If True, plot will be written to file and a value must be set for output_file
    :type save_plot:                bool

    :param date_field_name:         Name of date field in file
    :type date_field_name:          str

    :param time_field_name:         Name of time field in file
    :type time_field_name:          str

    :param elevation_field_name:    Name of elevation field in file
    :type elevation_field_name:     str

    :param x_padding:               Multiplier for maximum elevation to determine an ending interval for the x-axis.
                                    E.g., if max value is 100 and x_padding is 1.1 then the ending bound would be 110.
    :type x_padding:                float

    :param n_samples:               The number of samples to generate over the x-axis space
    :type n_samples:                int

    :param output_file:             Full path with file name and extension to an output file
    :type output_file:              str

    :param dpi:                     The resolution in dots per inch
    :type dpi:                      int

    :param style:                   Seaborn style designation
    :type style:                    str

    :param font_scale:              Scaling factor for font size
    :type font_scale:               float

    :param figsize:                 Tuple of figure size (x, y)
    :type figsize:                  Tuple[int]

    :param data_color:              Color of data line
    :type data_color:               str

    :param lognorm_color:           Color of data line
    :type lognorm_color:            str

    :param transparency             Alpha value from 0 to 1 for transparency of histogram
    :type transparency              float

    :param n_bins:                  The number of equal-width bins in the range.
    :type n_bins:                   int

    :param title                    Plot title.  If None, then no title will be displayed.
    :type title                     Union[str, None]

    """

    # prepare data for plotting
    plot = PlotGageData(data=data,
                        date_field_name=date_field_name,
                        time_field_name=time_field_name,
                        elevation_field_name=elevation_field_name,
                        style=style,
                        font_scale=font_scale,
                        figsize=figsize,
                        show_plot=show_plot,
                        save_plot=save_plot,
                        output_file=output_file,
                        dpi=dpi)

    # sort elevation value in ascending order
    z_sorted = plot.sorted_elevation()

    # generate an evenly spaced interval of values encompassing water surface elevation for use in scaling the x-axis
    x_data = plot.generate_wse_interval(x_padding=x_padding, n_samples=n_samples)

    # plot exceedance probability as a percentage
    exceedance_probability_series = plot.calculate_exceedance_probability()
    plot.ax.semilogx(exceedance_probability_series, z_sorted, ls='', marker='o', label='data', color=data_color)

    # plot cumulative distribution outputs
    cumulative = 100 * (1.0 - plot.calculate_cdf(data=x_data))
    plot.ax.plot(cumulative, x_data, label='lognormal', color=lognorm_color)

    # set interval spacing and formatting of the x-axis
    minor_locator = FixedLocator([1, 2, 5, 10, 20, 50, 100])
    plot.ax.xaxis.set_major_locator(minor_locator)
    plot.ax.xaxis.set_major_formatter(ScalarFormatter())
    plot.ax.xaxis.set_major_formatter(FormatStrFormatter("%d"))

    # set axis limits
    plot.ax.set_xlim(1, 100)
    plot.ax.set_ylim(0, plot.z_max)

    plot.ax.set_xlabel('Exceedance Probability (%)')
    plot.ax.set_ylabel('Water Surface Elevation (m)')
    plot.ax.invert_xaxis()
    plot.ax.legend(loc=0, framealpha=0.5)

    plt.title(title)

    # handle display or output options
    plot.output_handler()

    plt.close()


def plot_inundation_hectare_hours(data: Union[str, pd.DataFrame],
                                  show_plot: bool = True,
                                  save_plot: bool = False,
                                  output_file: Union[str, None] = None,
                                  dpi: int = 150,
                                  y_pad_fraction: float = 0.15,
                                  x_pad_fraction: float = 0.05,
                                  style: str = 'whitegrid',
                                  font_scale: float = 1.2,
                                  figsize: Tuple[int] = (12, 8),
                                  fill_color: str = '#005AB5',
                                  transparency: float = 0.7,
                                  title: Union[str, None] = 'Hectare Hours of Inundation'):
    """Plot of the hectare hours of inundation over water surface elevations.

    :param data:                    An data frame or CSV file containing inundation data as a result of
                                    atiim.simulate_inundation()
    :type data:                     Union[str, pd.DataFrame]

    :param show_plot:               If True, plot will be displayed
    :type show_plot:                bool

    :param save_plot:               If True, plot will be written to file and a value must be set for output_file
    :type save_plot:                bool

    :param output_file:             Full path with file name and extension to an output file
    :type output_file:              str

    :param dpi:                     The resolution in dots per inch
    :type dpi:                      int

    :param y_pad_fraction:          A decimal fraction of the maximum elevation value to use as a padding on the Y axis
    :type y_pad_fraction:           float

    :param x_pad_fraction:          A decimal fraction of the maximum hectare hour value to use as a padding on the
                                    X axis
    :type x_pad_fraction:           float

    :param style:                   Seaborn style designation
    :type style:                    str

    :param font_scale:              Scaling factor for font size
    :type font_scale:               float

    :param figsize:                 Tuple of figure size (x, y)
    :type figsize:                  Tuple[int]

    :param fill_color:              Color of filled area in plot
    :type fill_color:               str

    :param transparency             Alpha value from 0 to 1 for transparency
    :type transparency              float

    :param title                    Plot title.  If None, then no title will be displayed.
    :type title                     Union[str, None]

    """

    # prepare data for plotting
    plot = PlotInundationData(data=data,
                              style=style,
                              font_scale=font_scale,
                              figsize=figsize,
                              show_plot=show_plot,
                              save_plot=save_plot,
                              output_file=output_file,
                              dpi=dpi)

    # pad min and max Y values for axis
    z_max = plot.elevation.max()
    z_min = plot.elevation.min()
    y_padding = z_max * y_pad_fraction

    # pad max x axis value
    h_max = plot.hectare_hours.max()
    h_min = plot.hectare_hours.min()
    x_padding = h_max * x_pad_fraction

    # set axis limits
    plt.ylim(ymin=(z_min - y_padding),
             ymax=(z_max + y_padding))

    plt.xlim(xmin=h_min,
             xmax=h_max + x_padding)

    plt.plot(plot.hectare_hours, plot.elevation, 'black')

    # fill area under the curve
    plt.fill_betweenx(plot.elevation, plot.hectare_hours, color=fill_color, alpha=transparency)

    plt.title(title)
    plt.xlabel('Hectare Hours')
    plt.ylabel('Water Surface Elevation (m)')

    # handle display or output options
    plot.output_handler()

    plt.close()


def plot_inundation_perimeter(data: Union[str, pd.DataFrame],
                              show_plot: bool = True,
                              save_plot: bool = False,
                              output_file: Union[str, None] = None,
                              dpi: int = 150,
                              x_pad_fraction: float = 0.05,
                              style: str = 'whitegrid',
                              font_scale: float = 1.2,
                              figsize: Tuple[int] = (12, 8),
                              title: Union[str, None] = 'Inundation Perimeter by Water Surface Elevation'):
    """Plot perimeter of inundation through the water surface elevation time series.

    :param data:                    An data frame or CSV file containing inundation data as a result of
                                    atiim.simulate_inundation()
    :type data:                     Union[str, pd.DataFrame]

    :param show_plot:               If True, plot will be displayed
    :type show_plot:                bool

    :param save_plot:               If True, plot will be written to file and a value must be set for output_file
    :type save_plot:                bool

    :param output_file:             Full path with file name and extension to an output file
    :type output_file:              str

    :param dpi:                     The resolution in dots per inch
    :type dpi:                      int

    :param x_pad_fraction:          A decimal fraction of the maximum hectare hour value to use as a padding on the
                                    X axis
    :type x_pad_fraction:           float

    :param style:                   Seaborn style designation
    :type style:                    str

    :param font_scale:              Scaling factor for font size
    :type font_scale:               float

    :param figsize:                 Tuple of figure size (x, y)
    :type figsize:                  Tuple[int]

    :param title                    Plot title.  If None, then no title will be displayed.
    :type title                     Union[str, None]

    """

    # prepare data for plotting
    plot = PlotInundationData(data=data,
                              style=style,
                              font_scale=font_scale,
                              figsize=figsize,
                              show_plot=show_plot,
                              save_plot=save_plot,
                              output_file=output_file,
                              dpi=dpi)

    # plot perimeter line with points
    plt.plot(plot.perimeter, plot.elevation, 'ko', plot.perimeter, plot.elevation, 'k')

    # plot mean elevation line
    plt.axhline(plot.elevation.mean(), color='black', linestyle="--", label='_nolegend_')

    p_max = plot.perimeter.max()
    x_padding = p_max * x_pad_fraction
    plt.xlim(xmin=0, xmax=p_max + x_padding)

    plot.ax.set(ylabel='Water Surface Elevation (m)',
                xlabel='Inundation Perimeter (m)',
                title=title)

    # format the x-axis to show thousands as K (e.g., 15,000 as 15K)
    plot.set_xaxis_to_thousands()

    # handle display or output options
    plot.output_handler()

    plt.close()


def plot_inundation_area(data: Union[str, pd.DataFrame],
                         show_plot: bool = True,
                         save_plot: bool = False,
                         output_file: Union[str, None] = None,
                         dpi: int = 150,
                         style: str = 'whitegrid',
                         font_scale: float = 1.2,
                         figsize: Tuple[int] = (12, 8),
                         title: Union[str, None] = 'Inundation Area by Water Surface Elevation',
                         inflection_pt_size: int = 14):
    """Plot area of inundation through the water surface elevation time series.

    :param data:                    An data frame or CSV file containing inundation data as a result of
                                    atiim.simulate_inundation()
    :type data:                     Union[str, pd.DataFrame]

    :param show_plot:               If True, plot will be displayed
    :type show_plot:                bool

    :param save_plot:               If True, plot will be written to file and a value must be set for output_file
    :type save_plot:                bool

    :param output_file:             Full path with file name and extension to an output file
    :type output_file:              str

    :param dpi:                     The resolution in dots per inch
    :type dpi:                      int

    :param style:                   Seaborn style designation
    :type style:                    str

    :param font_scale:              Scaling factor for font size
    :type font_scale:               float

    :param figsize:                 Tuple of figure size (x, y)
    :type figsize:                  Tuple[int]

    :param title                    Plot title.  If None, then no title will be displayed.
    :type title                     Union[str, None]

    :param inflection_pt_size:      Size of the inflection point marker
    :type inflection_pt_size:       int

    """

    # prepare data for plotting
    plot = PlotInundationData(data=data,
                              style=style,
                              font_scale=font_scale,
                              figsize=figsize,
                              show_plot=show_plot,
                              save_plot=save_plot,
                              output_file=output_file,
                              dpi=dpi)

    # calculate the inflection point for bankfull elevation
    bankfull_elevation, bankfull_area = calculate_bankfull_elevation(df=plot.df)

    # plot line with points for inundated area by water surface elevation
    plt.plot(plot.area, plot.elevation, 'ko', plot.area, plot.elevation, 'k')

    # plot the point for bankfull elevation
    plt.plot(bankfull_area, bankfull_elevation, 'r^', ms=inflection_pt_size)

    plot.ax.set(ylabel='Water Surface Elevation (m)',
                xlabel='Area (m$^2$)',
                title=title)

    # format the x-axis to show thousands as K (e.g., 15,000 as 15K)
    plot.set_xaxis_to_thousands()

    # handle display or output options
    plot.output_handler()

    plt.close()


def plot_hypsometric_curve(data: Union[str, pd.DataFrame],
                           elevation_interval: float = 0.1,
                           min_elevation: Union[float, None] = None,
                           max_elevation: Union[float, None] = None,
                           show_plot: bool = True,
                           save_plot: bool = False,
                           output_file: Union[str, None] = None,
                           dpi: int = 150,
                           x_field_name: str = "dem_area_at_elevation",
                           y_field_name: str = "dem_elevation",
                           x_label: str = 'Area (m$^2$)',
                           y_label: str = 'Elevation (m)',
                           title: Union[str, None] = 'Hypsometric Curve',
                           style: str = 'whitegrid',
                           font_scale: float = 1.2,
                           figsize: Tuple[int] = (12, 8),
                           color: str = 'black'):
    """Plot a hypsometric curve as an elevation-area relationship assessment metric of the landform shape at a site.
    Provides basic metric of opportunity for inundation and habitat opportunity.

    :param data:                    Either a full path with file name and extension to the input DEM raster or
                                    a data frame processed using atiim.hypsometric_curve()
    :type data:                     Union[str, pd.Dataframe]

    :param elevation_interval:      Optional. Elevation sample spacing in the units of the input DEM. Set when using a
                                    DEM raster file input.
    :type elevation_interval:       float

    :param min_elevation:           Optional.  Minimum elevation to sample from.  Default is to use the minimum
                                    elevation of the raster as a starting point. Set when using a DEM raster file input.
    :type min_elevation:            float

    :param max_elevation:           Optional.  Maximum elevation to sample from.  Default is to use the maximum
                                    elevation of the raster as a starting point. Set when using a DEM raster file input.
    :type max_elevation:            float

    :param show_plot:               If True, plot will be displayed
    :type show_plot:                bool

    :param save_plot:               If True, plot will be written to file and a value must be set for output_file
    :type save_plot:                bool

    :param output_file:             Full path with file name and extension to an output file
    :type output_file:              str

    :param dpi:                     The resolution in dots per inch
    :type dpi:                      int

    :param x_field_name:            Field name of data for the x-axis
    :type x_field_name:             str

    :param y_field_name:            Field name of data for the y-axis
    :type y_field_name:             str

    :param x_label:                 Label for the x-axis
    :type x_label:                  str

    :param y_label:                 Label for the y-axis
    :type y_label:                  str

    :param title:                   Title for the plot if desired.  Use None for no title.
    :type title:                    Union[str, None]

    :param style:                   Seaborn style designation
    :type style:                    str

    :param font_scale:              Scaling factor for font size
    :type font_scale:               float

    :param figsize:                 Tuple of figure size (x, y)
    :type figsize:                  Tuple[int]

    :param color:                   Color of line and markers in plot
    :type color:                    str

    """

    # prepare data for plotting
    plot = PlotDem(data=data,
                   elevation_interval=elevation_interval,
                   min_elevation=min_elevation,
                   max_elevation=max_elevation,
                   style=style,
                   font_scale=font_scale,
                   figsize=figsize,
                   show_plot=show_plot,
                   save_plot=save_plot,
                   output_file=output_file,
                   dpi=dpi)

    # seaborn settings
    sns.set(style=style, font_scale=font_scale)

    sns.lineplot(x=x_field_name,
                 y=y_field_name,
                 marker='o',
                 data=plot.df,
                 color=color)

    plot.ax.set(ylabel=y_label,
                xlabel=x_label,
                title=title)

    # format the x-axis to show thousands as K (e.g., 15,000 as 15K)
    plot.set_xaxis_to_thousands()

    # handle display or output options
    plot.output_handler()

    plt.close()



from typing import Tuple

import numpy as np
import pandas as pd
from scipy.ndimage import gaussian_filter1d


def calculate_bankfull_elevation(df: pd.DataFrame,
                                 smooth_data: bool = False,
                                 smooth_sigma: int = 100,
                                 area_field_name: str = 'area',
                                 elevation_field_name: str = 'elevation') -> Tuple[float]:
    """Calculate the bankfull elevation point, or the first point of inflection, using the index of the
    change in sign of the second derivatives of the accumulating area with increasing elevation.  If the
    input data is noisy, you may choose to smooth the data using a Gaussian filter.

    :param df:                      Desingned for output from the atiim.simulate_inundation() function.  Though can
                                    be used with any data frame having an area and elevation field with data.
    :type df:                       pd.DataFrame

    :param smooth_data:             Optional.  Smooth noisy data using a Gaussian filter.  Use in combination with
                                    the smooth_sigma setting.
    :type smooth_data:              bool

    :param smooth_sigma:            Optional.  Standard deviation for Gaussian kernel.  Use when smooth_data is set
                                    to True.
    :type smooth_sigma:             int

    :param area_field_name:         Optional.  Name of area field in data frame.  Default:  'area'
    :type area_field_name:          str

    :param elevation_field_name:    Optional.  Name of elevation field in data frame.  Default 'elevation'
    :type elevation_field_name:     str

    :returns:                       [0] Bankfull elevation value.  First inflection point.
                                    [1] Bankfull area value

    """

    # smooth data using a Gaussian filter to remove noise if desired
    if smooth_data:
        area_data = gaussian_filter1d(df[area_field_name], sigma=smooth_sigma)
    else:
        area_data = df[area_field_name]

    # calculate the second derivatives
    second_derivatives = np.gradient(np.gradient(area_data))

    # get the index locations in the second derivative plot representing the sign change (a.k.a., inflection points)
    inflection_indices = np.where(np.diff(np.sign(second_derivatives)))[0]

    # drop the first value in the series if it shows up as an inflection point
    inflection_indices = inflection_indices[inflection_indices > 0]

    # bankfull elevation is determined by the first non-zero index inflection point
    bankfull_elevation = df[elevation_field_name].values[inflection_indices[0]]

    # bankfull area corresponding to the inflection point
    bankfull_area = df[area_field_name].values[inflection_indices[0]]

    return bankfull_elevation, bankfull_area



import os
import logging
import warnings
from typing import Union

import rasterio
import numpy as np
import pandas as pd
import geopandas as gpd
import rasterio.mask
import shapely.speedups
from rasterio.crs import CRS
from rasterio.features import shapes
from rasterio.transform import Affine
from joblib import delayed, Parallel

from .gage import process_gage_data
from .dem import create_basin_dem

# enable shapely speedups for topology operations
shapely.speedups.enable()


# TODO:  remove ignore warning generated for a future shapely depreciation in rasterio.features.shapes
warnings.filterwarnings('ignore')


def process_slice(arr: np.ndarray,
                  upper_elev: float,
                  gage_gdf: gpd.GeoDataFrame,
                  water_elev_freq: dict,
                  run_name: str,
                  hour_interval: float,
                  transform: Affine,
                  target_crs: CRS,
                  write_shapefile: bool = True,
                  output_directory: Union[str, None] = None) -> gpd.GeoDataFrame:
    """Create a water level polygon shapefile containing a single feature that represents
    the grid cells of an input DEM that are less than or equal to an upper elevation level.

    :param arr:                     2D array from raster band read
    :type arr:                      np.ndarray

    :param upper_elev:              Elevation value for the upper bound of the elevation interval
    :type upper_elev:               float

    :param gage_gdf:                GeoDataFrame of the gage location point
    :type gage_gdf:                 gpd.GeoDataFrame

    :param water_elev_freq:         Dictionary of water elevation frequency {elev:  frequency}
    :type water_elev_freq:          dict

    :param run_name:                Name of run, all lowercase and only underscore separated.
    :type run_name:                 str

    :param hour_interval:           Time step of inundation extent.  Either 1.0 or 0.5.
    :type hour_interval:            float

    :param transform:               Trasformation object from the rasterio source raster
    :type transform:                Affine

    :param target_crs:              Coordinate reference system (CRS) object from the rasterio source raster to
                                    be used in projecting the water level polygons.
    :type target_crs:               CRS

    :param write_shapefile:         Optional.  Choice to write the GeoDataFrame water level polygon as a shapefile.
                                    Set output directory if True.  Default is True.

    :type write_shapefile:          bool

    :param output_directory:        Full path to a write-enabled directory to write output files to if write_shapefile
                                    is set to True
    :type output_directory:         str

    :return:                        A geopandas data frame of a polygon intersecting the gage point location for
                                    the target elevation interval.

    """

    # TODO:  account for different units
    if hour_interval not in (1.0, 0.5):
        msg = f"The hour interval of '{hour_interval}' is not currently supported.  Please use either 1.0 or 0.5 (half hour)."
        raise AssertionError(msg)

    # generate a feature id from the elevation value
    feature_id = int(upper_elev * 100)

    # create every value greater than or equal to the upper elevation to 1, others to 0
    arx = np.where(arr <= upper_elev, 1, 0).astype(np.int16)

    # build each feature based on the extracted grid cells from the array
    results = list(
        {'properties': {'raster_val': val}, 'geometry': shp}
        for index, (shp, val) in enumerate(
            shapes(arx, mask=None, transform=transform))
    )

    # list of geometries
    geoms = list(results)

    # build geopandas dataframe from geometries
    gdf = gpd.GeoDataFrame.from_features(geoms, crs=target_crs)

    # only keep the ones
    gdf = gdf.loc[gdf['raster_val'] == 1]

    # only keep the polygon intersecting the gage
    gdf['valid'] = gdf.intersects(gage_gdf.geometry.values[0])
    gdf = gdf.loc[gdf['valid']].copy()

    # ensure at least one polygon intersects the gage
    if gdf.shape[0] == 0:
        msg = "Gage location point not aligned with valid elevation in DEM.  Relocate gage location point to fall within valid elevation."
        raise AssertionError(msg)

    # dissolve into a single polygon
    gdf = gdf.dissolve('raster_val')
    gdf.reset_index(inplace=True)

    # add fields
    gdf['id'] = feature_id
    gdf['frequency'] = water_elev_freq[round(upper_elev, 1)]
    gdf['elevation'] = upper_elev
    gdf['area'] = gdf.geometry.area
    gdf['hectares'] = gdf['area'] * 0.0001
    gdf['perimeter'] = gdf.geometry.length
    gdf['hect_hours'] = (gdf['frequency'] / hour_interval) * gdf['hectares']
    gdf['run_name'] = run_name

    # drop unneeded fields
    gdf.drop(columns=['raster_val', 'valid'], inplace=True)

    # write to file if desired
    if write_shapefile:

        if output_directory is None:
            msg = 'Please pass a value for output_directory if choosing to write shapefile outputs.'
            raise AssertionError(msg)

        out_file = os.path.join(output_directory, f'wl_{feature_id}_{run_name}.shp')
        gdf.to_file(out_file)

    return gdf


def simulate_inundation(dem_file: str,
                        basin_shp: str,
                        gage_shp: str,
                        gage_data_file: str,
                        run_name: str,
                        output_directory: Union[str, None] = None,
                        write_csv: bool = True,
                        elevation_interval: float = 0.1,
                        hour_interval: float = 1.0,
                        n_jobs: int = 1,
                        verbose: bool = False):
    """Worker function to simulate inundation over a DEM using a stable gage location and accompanying water level
    time series data.

    :param dem_file:                Full path with file name and extension to the input DEM raster file.
    :type dem_file:                 str

    :param basin_shp:               Full path with file name and extension to the target basin shapefile
    :type basin_shp:                str

    :param gage_shp:                Full path with file name and extension to the target gage shapefile
    :type gage_shp:                 str

    :param gage_data_file:          Full path with file name and extension to the gage data file.
    :type gage_data_file:           str

    :param output_directory:        Full path to a write-enabled directory to write output files to
    :type output_directory:         str

    :param run_name:                Name of run, all lowercase and only underscore separated.
    :type run_name:                 str

    :param write_csv:               Choice to write a CSV file of inundation metric outputs
    :type write_csv:                bool

    :param elevation_interval:      Step for elevation to be processed.
    :type elevation_interval:       float

    :param hour_interval:           Time step of inundation extent.  Either 1.0 or 0.5.
    :type hour_interval:            float

    :param n_jobs:                  The maximum number of concurrently running jobs, such as the number of Python
                                    worker processes when backend=”multiprocessing” or the size of the thread-pool
                                    when backend=”threading”. If -1 all CPUs are used. If 1 is given, no parallel
                                    computing code is used at all, which is useful for debugging. For n_jobs
                                    below -1, (n_cpus + 1 + n_jobs) are used. Thus for n_jobs = -2, all
                                    CPUs but one are used. None is a marker for ‘unset’ that will be interpreted
                                    as n_jobs=1 (sequential execution) unless the call is performed under
                                    a parallel_backend context manager that sets another value for n_jobs.
    :type n_jobs:                   int

    :param verbose:                 Choice to log verbosely
    :type verbose:                  bool

    """

    # process gage data file
    min_gage_elev, max_gage_elev, water_elev_freq = process_gage_data(gage_data_file, verbose=verbose)

    # read in gage shapefile to a geodataframe
    gage_gdf = gpd.read_file(gage_shp)

    with rasterio.Env():

        # clip the input DEM to a target basin contributing area
        masked_dem_file = create_basin_dem(basin_shp, dem_file, run_name, output_directory)

        with rasterio.open(masked_dem_file) as src:

            # read the raster band into a number array
            arr = src.read(1)

            # convert the raster nodata value to numpy nan
            arr[arr == src.nodata] = np.nan

            raster_min = float(np.nanmin(arr))
            raster_max = float(np.nanmax(arr))

            # use the minimum bounding elevation e.g., the max of min available
            elev_min = max([min_gage_elev, raster_min])

            # use the maximum bounding elevation e.g., the min of max available
            elev_max = min([max_gage_elev, raster_max])

            # construct elevation upper bounds to process for each slice
            elev_slices = np.arange(elev_min, elev_max + elevation_interval, elevation_interval)

            # process all elevation slices in parallel
            feature_list = Parallel(n_jobs=n_jobs)(
                delayed(process_slice)(arr=arr,
                                       upper_elev=upper_elev,
                                       output_directory=output_directory,
                                       gage_gdf=gage_gdf,
                                       water_elev_freq=water_elev_freq,
                                       run_name=run_name,
                                       hour_interval=hour_interval,
                                       transform=src.transform,
                                       target_crs=src.crs)
                for upper_elev in elev_slices)

            # concatenate individual GeoDataFrames into a single frame
            result_df = pd.concat(feature_list)

            # drop geometry
            result_df.drop(columns=['geometry'], inplace=True)

            # write data frame to file removing geometry if so desired
            if write_csv:

                if output_directory is None:
                    msg = 'Please pass a value for output_directory if choosing to write shapefile outputs.'
                    raise AssertionError(msg)

                out_file = os.path.join(output_directory, f'inundation_metrics_{run_name}.csv')
                result_df.to_csv(out_file, index=False)

            if verbose:
                logging.info(f"Minimum DEM Elevation:  {round(raster_min, 2)}")
                logging.info(f"Maximum DEM Elevation:  {round(raster_max, 2)}")
                logging.info(f"Bounded DEM Elevation:  {round(min(elev_slices), 2)}")
                logging.info(f"Bounded DEM Elevation:  {round(max(elev_slices), 2)}")

            return pd.DataFrame(result_df)



from .gage import import_gage_data, process_gage_data
from .hypsometric import hypsometric_curve
from .inundation import simulate_inundation
from .plots import plot_wse_timeseries, plot_wse_cumulative_distribution, plot_wse_probability_density, \
    plot_wse_exceedance_probability, plot_inundation_hectare_hours, plot_inundation_perimeter, plot_inundation_area, \
    plot_hypsometric_curve

from .package_data import SampleData


__version__ = "0.0.0"



import pkg_resources


class SampleData:
    """ATIIM sample data."""

    @property
    def sample_gage_data_file(self) -> str:
        """Retrieve the sample gage data containing a time series of water surface elevation.

        :return:                Full path with file name and extension to the sample file

        """

        return pkg_resources.resource_filename('atiim', 'data/water_level.csv')

    @property
    def sample_gage_shapefile(self) -> str:
        """Retrieve the sample gage location shapefile.

        :return:                Full path with file name and extension to the sample file

        """

        return pkg_resources.resource_filename('atiim', 'data/gage_location_1.shp')

    @property
    def sample_basin_shapefile(self) -> str:
        """Retrieve the sample basin shapefile.

        :return:                Full path with file name and extension to the sample file

        """

        return pkg_resources.resource_filename('atiim', 'data/basin_1.shp')

    @property
    def sample_dem(self) -> str:
        """Retrieve the sample DEM.

        :return:                Full path with file name and extension to the sample file

        """

        return pkg_resources.resource_filename('atiim', 'data/dem.sdat')



import re
from setuptools import setup, find_packages


def readme():
    """Return the contents of the project README file."""
    with open('README.md') as f:
        return f.read()


version = re.search(r"__version__ = ['\"]([^'\"]*)['\"]", open('atiim/__init__.py').read(), re.M).group(1)

setup(
    name='atiim',
    version=version,
    packages=find_packages(),
    url='https://github.com/crvernon/atiim',
    license='BSD-2-Clause',
    author='André M. Coleman; Chris R. Vernon',
    description='The Area Time Inundation Index Model',
    long_description=readme(),
    long_description_content_type="text/markdown",
    python_requires='>=3.7.*, <4',
    include_package_data=True,
    install_requires=[
        'numpy>=1.19.4',
        'pandas>=1.1.4',
        'rasterio>=1.2.3',
        'requests>=2.25.1',
        'joblib>=1.0.1',
        'matplotlib>=3.3.3',
        'seaborn>=0.11.1',
        'fiona>=1.8.19',
        'pyproj>=3.0.1',
        'rtree>=0.9.7',
        'shapely>=1.7.1',
        'geopandas>=0.9.0'
    ],
    extras_require={
        'dev': [
            'build>=0.5.1',
            'nbsphinx>=0.8.6',
            'setuptools>=57.0.0',
            'sphinx>=4.0.2',
            'sphinx-panels>=0.6.0',
            'sphinx-rtd-theme>=0.5.2',
            'twine>=3.4.1',
            'pytest>=6.2.4',
            'pytest-cov>=2.12.1'
        ]
    }
)
